@@ -38,7 +38,7 @@ class SankeyNode:
         @param color : any matplotlib color specifier
         @param label : alternative name to display. If not label, uses `name`
         @param label_format : format string that can optionally have 'label' and 'value' fields. Set to '' to remove labels
-        @param label_pos : position of the label {'left', 'top'}
+        @param label_pos : position of the label {'left', 'right', 'top', 'bottom', 'center'}
         @param label_pad_x : padding in x from the node, as fraction of distance between nodes
         @param label_opts : dictionary with additional options for the label. See matplotlib.axis.text() for options
         '''
@@ -63,9 +63,9 @@ class SankeyNode:
         self.text = None
 
 
-    def get_flow_y(self, i, side='inflows'):
+    def get_flow_y(self, i, side):
         '''
-        Returns the (y_low, y_hi) position of flow i on the given side
+        Returns the (y_low, y_hi) position of flow i on the given side {'inflows', 'outflows'}
         '''
         flows = getattr(self, side)
         if self.align_y == 'top':
@@ -93,13 +93,28 @@ class SankeyNode:
                 y = self.y + self.height / 2
                 ha = 'right'
                 va = 'center'
+            elif self.label_pos == 'right':
+                x = self.x + self.width + self.label_pad_x
+                y = self.y + self.height / 2
+                ha = 'left'
+                va = 'center'
             elif self.label_pos == 'top':
                 x = self.x + self.width / 2
                 y = self.y + self.height
                 ha = 'center'
                 va = 'bottom'
+            elif self.label_pos == 'bottom':
+                x = self.x + self.width / 2
+                y = self.y
+                ha = 'center'
+                va = 'top'
+            elif self.label_pos == 'center':
+                x = self.x + self.width / 2
+                y = self.y + self.height / 2
+                ha = 'center'
+                va = 'center'
             else:
-                raise NotImplementedError("label_pos: " + self.label_pos)
+                raise ValueError("label_pos: " + self.label_pos)
             self.text = ax.text(x, y, self.label_format.format(label=self.label, value=self.value), ha=ha, va=va, **self.label_opts)
 
 
@@ -196,17 +211,19 @@ class Sankey:
 
     def __init__(self, flows=None, nodes=None, align_y='top', 
                     cmap=plt.cm.tab10, flow_color_mode='dest', flow_color_mode_alpha=0.6, 
-                    node_width=0.03, node_pad_y_min=0.01, node_pad_y_max=0.05,
+                    node_width=0.03, node_height_pad_min=0, node_pad_y_min=0.01, node_pad_y_max=0.05,
                     node_opts={}, flow_opts={}, **kwargs):
         '''
         @param flows : see Sankey.sankey(). Can optionally input flows here as a shortcut
         @param nodes : see Sankey.sankey(). Can optionally input nodes here as a shortcut
+        @param align_y : vertical alignment of the nodes {'top', 'center', 'bottom', 'justify', 'tree', 'tree clamp'}
         @param cmap : colormap for default colors
         @param flow_color_mode : {'source', 'dest', 'lesser', 'greater', None}
                     'source', 'dest' : the flows will be colored the same as the source/destination node
                     'lesser', 'greater' : the flows will be colored as the node with the lesser/greater value 
         @param flow_color_mode_alpha : transparency of the flows
         @param node_width : width of the nodes, as fraction of distance between nodes
+        @param node_height_pad_min : if node.height < node_height_pad_min, adds additional y padding. Value is a fraction of axis height.
         @param node_pad_y_min : minimum vertical padding between nodes, as fraction of distance between nodes
         @param node_pad_y_max : maximum vertical padding between nodes, as fraction of distance between nodes
         @param node_opts : a dictionary of options to SankeyNode() that is applied to all nodes
@@ -217,6 +234,7 @@ class Sankey:
         self.flow_color_mode = flow_color_mode
         self.flow_color_mode_alpha = flow_color_mode_alpha
         self.node_width = node_width
+        self.node_height_pad_min = node_height_pad_min
         self.node_pad_y_min = node_pad_y_min
         self.node_pad_y_max = node_pad_y_max
         self.node_opts = node_opts
@@ -271,17 +289,83 @@ class Sankey:
 
         return arr
 
-    
-    def _level_node_padding(self, level_value, level_n):
-        '''
-        @param level_value : Total value in this level, AFTER scaling
-        @returns the padding to use in this level
-        '''
-        if level_n < 2:
-            return 0
-        max_padding = (1 - level_value) / (level_n - 1)
-        return max(self.node_pad_y_min, min(max_padding, self.node_pad_y_max))
 
+    def _value_scale_level(self, nodes_level):
+        '''
+        Finds the value scale such that this level ranges exactly from y=[0,1] including padding
+        '''
+        min_pad = self.node_pad_y_min * (len(nodes_level) - 1)
+        heights = np.array([node[1] for node in nodes_level])
+        scale = np.sum(heights) / (1 - min_pad)
+        if self.node_height_pad_min == 0:
+            return scale
+        else:
+            if self.node_height_pad_min * len(nodes_level) + min_pad > 1:
+                raise ValueError('sankey: node_height_pad_min {} is too large for level with size {}'.format(self.node_height_pad_min, len(nodes_level)))
+            
+            nsmall = 0
+            while (smalls := heights / scale <= self.node_height_pad_min).any(): # any smalls left over = we need to increase the scale
+                nsmall += np.count_nonzero(smalls)
+                heights = heights[~smalls]
+                scale = np.sum(heights) / (1 - min_pad - nsmall * self.node_height_pad_min)
+
+            return scale
+
+    def _level_node_max_padding(self, nodes, value_scale):
+        '''
+        @returns the max padding that can be used in this level
+        '''
+        if len(nodes) < 2:
+            return 0
+        heights = np.array([node[1] for node in nodes]) / value_scale
+        level_value = np.sum(np.maximum(heights, self.node_height_pad_min))
+        return (1 - level_value) / (len(nodes) - 1)
+
+    def _get_node_ys(self, nodes, value_scale):
+        '''
+        Returns the y-positions of the nodes in a single level as (y_low, height)
+        '''
+        # Get padding
+        max_padding = self._level_node_max_padding(nodes, value_scale)
+        if self.align_y == 'justify':
+            node_pad_y = max_padding
+        else:
+            node_pad_y = min(max_padding, self.node_pad_y_max)
+
+        # Special case for n=1 with align = center or justify. align = top or bottom can use the general code below
+        if len(nodes) == 1 and self.align_y in ['center', 'justify']:
+            height = nodes[0][1] / value_scale
+            pad = (1 - height) / 2
+            return [(pad, height)]
+
+        # General case
+        ys = []
+        if self.align_y == 'bottom':
+            y = 0
+            for node in reversed(nodes):
+                height = node[1] / value_scale
+                if height < self.node_height_pad_min:
+                    pos = y + (self.node_height_pad_min - height) / 2
+                    ys.insert(0, (pos, height))
+                    y += self.node_height_pad_min + node_pad_y
+                else:
+                    ys.insert(0, (y, height))
+                    y += actual_height + node_pad_y
+        else:
+            y = 1
+            if self.align_y == 'center':
+                y -= (max_padding - node_pad_y) * (len(nodes) - 1) / 2
+            for node in nodes:
+                height = node[1] / value_scale
+                y -= height
+                if height < self.node_height_pad_min:
+                    y -= (self.node_height_pad_min - height) / 2
+                    ys.append((y, height))
+                    y -= (self.node_height_pad_min - height) / 2
+                else:
+                    ys.append((y, height))
+                y -= node_pad_y
+        return ys
 
     def sankey(self, flows, nodes=None):
         '''
@@ -295,34 +379,35 @@ class Sankey:
         if not nodes:
             nodes = self.infer_nodes(flows)
 
-        # Get scaling
-        _total_value_per_level = [sum(node[1] for node in level) for level in nodes]
-        max_level_value = np.max(_total_value_per_level)
-        max_level_n = np.max([len(level) for level,val in zip(nodes,_total_value_per_level) if max_level_value - val < 1e-5])
-        value_scale = max_level_value / (1 - self.node_pad_y_min * (max_level_n - 1))
+        # Get scaling. The widest level should extend exactly from y = 0 to 1 including padding
+        widest_level = 0
+        value_scale = 0
+        for i,nodes_level in enumerate(nodes):
+            scale = self._value_scale_level(nodes_level)
+            if scale > value_scale:
+                value_scale = scale
+                widest_level = i
 
         # Nodes
         self.nodes = []
         i_color = 0 # for automatic coloring
-        for level,node_level in enumerate(nodes):
+        for level,nodes_level in enumerate(nodes):
             arr = [] # nodes in this level
-            node_pad_y = self._level_node_padding(_total_value_per_level[level] / value_scale, len(node_level))
-            if self.align_y == 'top':
-                y = 1
-                for node in node_level:
-                    if node[1] <= 0:
-                        raise ValueError("Node value <= 0: {}".format(node))
-                    height = node[1] / value_scale 
-                    args = dict(color=self.cmap(i_color % self.cmap.N))
-                    args.update(self.node_opts)
-                    if len(node) > 2:
-                        args.update(node[2])
-                    arr.append(SankeyNode(level, y - height, self.node_width, height, node[0], node[1], **args))
-                    i_color += 1
-                    y -= height + node_pad_y
-            else:
-                raise NotImplementedError("align_y: " + align_y)
-            
+            ys = self._get_node_ys(nodes_level, value_scale)
+            for node,(y,height) in zip(nodes_level, ys):
+                # Error check
+                if node[1] < 0:
+                    print('Warning: Node has value < 0: {}'.format(node))
+
+                # Node configuration
+                args = dict(color=self.cmap(i_color % self.cmap.N))
+                args.update(self.node_opts) # global options
+                if len(node) > 2:
+                    args.update(node[2]) # individual options
+
+                # Create node
+                arr.append(SankeyNode(level, y, self.node_width, height, node[0], node[1], **args))
+                i_color += 1
             self.nodes.append(arr)
 
         # Flows
@@ -333,11 +418,11 @@ class Sankey:
 
             # Error check
             if not src:
-                raise KeyError("Bad flow - couldn't find node: {}".format(flow[0]))
+                raise KeyError("Bad flow - couldn't find souce node: {}".format(flow))
             if not des:
-                raise KeyError("Bad flow - couldn't find node: {}".format(flow[1]))
-            if flow[2] > src.value or flow[2] > des.value or flow[2] <= 0:
-                raise ValueError("Bad flow - bad weight: {}".format(flow[2]))
+                raise KeyError("Bad flow - couldn't find destination node: {}".format(flow))
+            if flow[2] > src.value or flow[2] > des.value or flow[2] < 0:
+                print("Warning: Bad flow - bad weight: {}".format(flow))
             if des_level <= src_level:
                 raise ValueError("Bad flow - flow is backwards: {}".format(flow))
 
@@ -363,6 +448,79 @@ class Sankey:
             args.update(custom_opts)
             self.flows.append(SankeyFlow(src, des, flow[2], **args))
 
+        # Post-creation layout
+        if 'tree' in self.align_y:
+            self._layout_tree(widest_level)
+    
+
+    def _layout_tree_level(self, nodes_level, right):
+        '''
+        @param right : This level is right of the max level, and will base the ideal positions off of the nodes' parents. Otherwise, uses the children.
+        '''
+        # Get the ideal y position such that the flows are horizontal. For nodes with a single parent,
+        # this is simply the y of the flow. For nodes with multiple parents, use a weighted average.
+        y_ideal = np.zeros(len(nodes_level))
+        has_stress = np.ones(len(nodes_level), dtype=bool)
+        for i,node in enumerate(nodes_level):
+            node.y = 0 # needed when we call get_flow_y below
+            total_ys = 0
+            total_weights = 0
+            for flow in node.inflows if right else node.outflows:
+                if right:
+                    y_anchor = flow.src.get_flow_y(flow.src_i, "outflows")[0]
+                    y_flow_node = node.get_flow_y(flow.des_i, "inflows")[0] # since we set y to 0 above, this is the position of the flow in the node
+                else:
+                    y_anchor = flow.des.get_flow_y(flow.des_i, "inflows")[0]
+                    y_flow_node = node.get_flow_y(flow.src_i, "outflows")[0] # since we set y to 0 above, this is the position of the flow in the node
+                y_ideal_node = np.clip(y_anchor - y_flow_node, 0, 1)
+                total_ys += y_ideal_node * flow.value
+                total_weights += flow.value
+            if total_weights > 0:
+                y_ideal[i] = total_ys / total_weights
+            else:
+                has_stress[i] = False
+
+        # Get the maximum height each node can have
+        y_max = np.zeros(len(nodes_level))
+        y = 1
+        for i,node in enumerate(nodes_level):
+            if node.height < self.node_height_pad_min:
+                y_max[i] = y - (self.node_height_pad_min + node.height) / 2
+                y -= self.node_height_pad_min + self.node_pad_y_min
+            else:
+                y_max[i] = y - node.height
+                y -= node.height + self.node_pad_y_min
+        y_flex = y_max[-1] # amount of extra y space available
+
+        # Start the nodes from their maximum (top) position, then calculate the amount of "stress" each node
+        # has to reach its ideal position. Large stress = wants to move down, negative stress = wants to move up.
+        # If a single node wants to push down but the ones below it want to move up, we want to average
+        # the stresses such that no node is too selfish. So use a weighted average of all nodes below it.
+        y_stress = (y_max - y_ideal) * has_stress
+        average_stress = (np.cumsum(y_stress[::-1]) / np.cumsum(has_stress[::-1]))[::-1]
+        y_desired_shift = np.minimum(y_stress, average_stress)
+        if 'clamp' in self.align_y: # we can clip here to ensure height of plot = 1 = widest layer, but not clipping looks fine too
+            y_desired_shift = np.clip(y_desired_shift, 0, y_flex) 
+        y_shift = np.maximum.accumulate(y_desired_shift) # if one node pushes down, the subsequent nodes must follow by at least as much
+        y_new = y_max - y_shift
+
+        # Set the new positions
+        for node,pos in zip(nodes_level, y_new):
+            node.y = pos
+
+    def _layout_tree(self, max_level):
+        '''
+        Adjusts the positions of self.nodes to group sibling nodes next to each other and their parent. Works best when the nodes form
+        a tree structure. This should be called after the flows are created but before draw(). 
+
+        The algorithm starts from max_level and works its way out in both directions (TODO). Note that the ordering of the nodes
+        (in top-to-bottom order) always takes precedence.  
+        '''
+        for nodes_level in self.nodes[max_level + 1:]:
+            self._layout_tree_level(nodes_level, True)
+        for nodes_level in reversed(self.nodes[:max_level]):
+            self._layout_tree_level(nodes_level, False)
+
     def draw(self, ax=None):
         # Get axes
         if ax is None:
